/*************************************************************************\
*                  Copyright (C) Jan Murawski, 2012.                      *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Public License as published                  *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
\*************************************************************************/

#include "TunnelClient.h"

#include <stdio.h>
#include <netdb.h>
#include <string.h>
#include <unistd.h>
#include <limits.h>
#include <linux/ip.h>
#include <linux/icmp.h>
#include <linux/if_ether.h>

TunnelClient::TunnelClient() 
{
    this->start();
}


TunnelClient::~TunnelClient() 
{
    close(tunnelSock);
    close(rawSockIn);
    close(rawSockOut);
}

void TunnelClient::execute()
{
    //
    // raw socket (caputer ehternet frames, strip ethernet header, only ip packets)
    //
    if ((rawSockIn = socket(PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP))) == -1)
    //if ((rawSockIn = socket(PF_PACKET, SOCK_DGRAM, IPPROTO_TCP)) == -1)
    {
        Logger::error("ERROR opening raw socket");
    }
    
    //
    // raw socket for sending (IPPROTO_RAW == sending only)
    //
    if ((rawSockOut = socket(PF_INET, SOCK_RAW, IPPROTO_RAW)) == -1)
    {
        Logger::error("ERROR opening raw socket");
    }
    
    //
    // connect tunnel
    //
    
    while(true)
   {
        // create socket
        tunnelSock = socket(AF_INET, SOCK_STREAM, 0);
        if (tunnelSock < 0) 
            Logger::error("ERROR opening socket");
    
	// assemble address
	struct sockaddr_in serv_addr;
	struct hostent *server;
	server = gethostbyname("127.0.0.1");
	bzero((char *) &serv_addr, sizeof(serv_addr));
	serv_addr.sin_family = AF_INET;
	bcopy((char *)server->h_addr, 
	(char *)&serv_addr.sin_addr.s_addr,
	server->h_length);
	serv_addr.sin_port = htons(1337);
//        serv_addr.sin_port = htons(1335);

	// connect to server
        Logger::log("Initiating tunnel\n");
	if (connect(tunnelSock,(struct sockaddr *) &serv_addr,sizeof(serv_addr)) < 0) 
		Logger::error("ERROR connecting");

        int bytesAvailableTunnel = 0;
        int bytesReadTunnel = 0; 
        char bufferRaw[USHRT_MAX];
        char bufferTunnel[USHRT_MAX];
        unsigned short ipSizeTunnel = 0;
        unsigned long localSubnetMin = _LOCAL_SUBNET_MIN_;
        unsigned long localSubnetMax = _LOCAL_SUBNET_MAX_;
        unsigned long allowedClients[_ALLOWED_CLIENTS_COUNT_] = _ALLOWED_CLIENTS_;
        unsigned long directIPs[_DIRECT_COUNT_] = _DIRECT_IPS_;
   
        while( bytesAvailableTunnel >= 0 && !socket_is_closed(tunnelSock) )
        {
            bool action = false;

            /*
            *  handle packets from tunnel
            */
            bytesAvailableTunnel = bytes_on_socket(tunnelSock);

            // read first 4 bytes
            // byte 3 and 4 are size of ip-packet
            if( bytesReadTunnel == 0 && bytesAvailableTunnel >= 4 )
            {
                bytesReadTunnel = read(tunnelSock, bufferTunnel, 4);

                if( bytesReadTunnel != 4 )
                    break;

                // grab packet size from ip header
                ipSizeTunnel = ntohs( *(unsigned short*)(bufferTunnel +2));
            }
            else if( bytesReadTunnel == 0 && socket_is_closed(tunnelSock) )
                break;

            // size from header has been read? 
            // read rest of packet as soon as it is fully available
            if( bytesReadTunnel == 4 && bytesAvailableTunnel +4 >= ipSizeTunnel  )
            {
                read(tunnelSock, bufferTunnel+4, ipSizeTunnel -4 );
                bytesReadTunnel = 0;

                struct iphdr* ip = (struct iphdr*)&bufferTunnel;
                //printf("0x%x :: 0x%x -- 0x%x  reply\n", ntohs(ip->check), ip->saddr, ip->daddr);

                if( basic_header_fields_valild(bufferTunnel, ipSizeTunnel) == false )
                    break;

                /*
                * send packet to local net
                */
                Logger::log("local put\n", Logger::VERBOSE);
                struct sockaddr_in connection;
                connection.sin_family = AF_INET;
                connection.sin_addr.s_addr = ip->daddr;

                sendto(rawSockOut, 
                    bufferTunnel, 
                    ipSizeTunnel, 
                    0, 
                    (struct sockaddr *)&connection, 
                    sizeof(struct sockaddr) );

                action = true;
            }
            else if( bytesReadTunnel == 4 && socket_is_closed(tunnelSock) )
                break;

            /*
            * handle packets from raw
            */
            int bytesAvailableRaw = bytes_on_socket(rawSockIn);
            if( bytesAvailableRaw > 0 )
            {
                // read one packet
                int bytesRead = read(rawSockIn, bufferRaw, USHRT_MAX);

                if( basic_header_fields_valild(bufferRaw, bytesRead) == false )
                    continue;
                struct iphdr* ip = (struct iphdr*)&bufferRaw;
                bytesRead = ntohs(ip->tot_len); // Win7 fix
                // Win7 sends packets whith empty trailing bytes
                // discard them

                // target is required to be NOT on local subnet
                unsigned long dstAddr = ntohl( ip->daddr );
                if( dstAddr >= localSubnetMin && dstAddr <= localSubnetMax )
                        continue;
                
                // source is required to be allowed
                unsigned long srcAddr = ntohl( ip->saddr );
                bool match = false;
                for(int i=0; i < _ALLOWED_CLIENTS_COUNT_; i++)
                {
                    if( allowedClients[i] == srcAddr )
                        match = true;
                }
                if( !match )
                    continue;
                
                // Packet should take direct way? (via proxy for example)
                match = false;
                for(int i=0; i < _DIRECT_COUNT_; i++)
                {
                    if( directIPs[i] == dstAddr )
                        match = true; 
                }
                if( match )
                {
                    /*
                    * send packet on direct path
                    */
                    struct sockaddr_in connection;
                    connection.sin_family = AF_INET;
                    connection.sin_addr.s_addr = ip->daddr;

                    sendto(rawSockOut, 
                        bufferRaw, 
                        bytesRead, 
                        0, 
                        (struct sockaddr *)&connection, 
                        sizeof(struct sockaddr) );
                    continue;
                }

                //printf("check: 0x%x src: 0x%x dst: 0x%x\n", ntohs(ip->check), 
                    //   ntohl(ip->saddr),
                    // ntohl(ip->daddr));

                // ICMP
                if(ip->protocol == IPPROTO_ICMP)
                {
                    // in case of icmp only allow echo requests
                    struct icmphdr* icmp = (struct icmphdr*) (bufferRaw + sizeof(struct iphdr));
                    if( icmp->type != ICMP_ECHO )
                        continue;
                }
                else if(ip->protocol == IPPROTO_TCP || ip->protocol == IPPROTO_UDP)
                {
                    // no further checks required
                    // grab all tcp and udp traffic
                }
                else
                    continue;

                // write packet to tunnel
                Logger::log("local grab\n", Logger::VERBOSE);
                int bytesWritten = write(tunnelSock, bufferRaw, bytesRead);
                // quit, if tunnel is closed
                if( bytesWritten != bytesRead )
                    break;

                action = true;
            }

            if( !action )
                usleep(1);
        }
        
        close(tunnelSock);
        Logger::log("Connection reset\n");
        sleep(10);
    }
    

   
    
    
}
