/*************************************************************************\
*                  Copyright (C) Jan Murawski, 2012.                      *
*                                                                         *
* This program is free software. You may use, modify, and redistribute it *
* under the terms of the GNU Public License as published                  *
* by the Free Software Foundation, either version 3 or (at your option)   *
* any later version. This program is distributed without any warranty.    *
\*************************************************************************/

#include "helper.h"

#include <asm/types.h>
#include <linux/ip.h>

#include <netinet/tcp.h>
#include <string.h>

long bytes_on_socket(int socket)
{
        size_t nbytes = 0;
        if ( ioctl(socket, FIONREAD, (char*)&nbytes) < 0 )  
        {
                return -1;
        }
        return( (long)nbytes );
}



bool socket_is_closed(int sock)
{
    struct pollfd fds;
    fds.fd = sock;
    fds.events = POLLRDHUP | POLLERR | POLLHUP | POLLNVAL;
    if( poll(&fds, 1, 0) < 0 )
        return true;

    if( fds.revents != 0 )
        return true;
    
    return false;
}


bool basic_header_fields_valild(char* bufferRaw, int bytesRead)
{
    struct iphdr* ip = (struct iphdr*)bufferRaw;
    
    if( bytesRead < 20 )
        return false;
    
    // ipv4 ?
    if( ( *bufferRaw & 0xf0) != 0x40 )
        return false;
   
    // protocol is ICMP, TCP, UDP ?
    if( ip->protocol != IPPROTO_UDP &&
            ip->protocol != IPPROTO_TCP && 
            ip->protocol != IPPROTO_ICMP )
        return false;

   // no broadcast?
   unsigned long dstAddr =  ip->daddr;
   //if( (dstAddr & 0xff000000) == 0xff000000 )
   if( *((unsigned char*)(&dstAddr)+3) == 0xff )
       return false;

   // no loopback?
   if( dstAddr == 0x7f000001 )
       return false;
   
   // seems to be valid length?
   if(ntohs(ip->tot_len) > bytesRead )  // Win7 sometimes sets ip->tot_len 6 bytes to large oO
   {
       return false;
   }
   
   return true;
}

long ip_checksum(unsigned short *addr, unsigned int count, long initial_sum) {
          /* Compute Internet Checksum for "count" bytes
            *         beginning at location "addr".
            */
       register long sum = initial_sum;


        while( count > 1 )  {
           /*  This is the inner loop */
               sum += * addr++;
               count -= 2;
       }
           /*  Add left-over byte, if any */
       if( count > 0 )
               sum += * (unsigned char *) addr;

           /*  Fold 32-bit sum to 16 bits */
       while (sum>>16)
           sum = (sum & 0xffff) + (sum >> 16);

       return ~sum;
}

struct tcp_pseudo /*the tcp pseudo header*/
{
  __u32 src_addr;
  __u32 dst_addr;
  __u8 zero;
  __u8 proto;
  __u16 length;
};

long get_tcp_checksum(struct iphdr * myip, struct tcphdr * mytcp) 
{
        u16 total_len = ntohs(myip->tot_len);

        int tcpopt_len = mytcp->doff*4 - 20;
        int tcpdatalen = total_len - (mytcp->doff*4) - (myip->ihl*4);
        int totaltcp_len = sizeof(struct tcphdr) + tcpopt_len + tcpdatalen;
        
        struct tcp_pseudo pseudohead;
        pseudohead.src_addr=myip->saddr;
        pseudohead.dst_addr=myip->daddr;
        pseudohead.zero=0;
        pseudohead.proto=IPPROTO_TCP;
        pseudohead.length=htons(sizeof(struct tcphdr) + tcpopt_len + tcpdatalen);
        
        unsigned short *ptr = (unsigned short*)&pseudohead;
        long sum = *ptr + *(ptr+1) + *(ptr+2) + *(ptr+3) + *(ptr+4) + *(ptr+5);

        return ip_checksum((unsigned short *)mytcp,
                totaltcp_len, 
                sum);
}

long get_udp_checksum(struct iphdr * myip, struct udphdr * udp) 
{
        struct tcp_pseudo pseudohead;
        pseudohead.src_addr=myip->saddr;
        pseudohead.dst_addr=myip->daddr;
        pseudohead.zero=0;
        pseudohead.proto=IPPROTO_UDP;
        pseudohead.length=udp->len;
        
        unsigned short *ptr = (unsigned short*)&pseudohead;
        long sum = *ptr + *(ptr+1) + *(ptr+2) + *(ptr+3) + *(ptr+4) + *(ptr+5);

        return ip_checksum((unsigned short *)udp,
                ntohs(udp->len), 
                sum);
}


